@@ -27,6 +27,7 @@
 
 #include "datatypes.h"
 #include "ch_conf.h"
+#include "ch_domain.h"
 #include "ch_events.h"
 #include "ch_interface.h"
 #include "ch_monitor.h"
@@ -37,6 +38,7 @@
 #include "virfile.h"
 #include "virjson.h"
 #include "virlog.h"
+#include "virpidfile.h"
 #include "virstring.h"
 
 #define VIR_FROM_THIS VIR_FROM_CH
@@ -131,28 +133,50 @@ virCHMonitorBuildPayloadJson(virJSONValue *content, virDomainDef *vmdef)
 {
     g_autoptr(virJSONValue) payload = virJSONValueNewObject();
 
-
     if (vmdef->os.kernel == NULL) {
-        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
-                       _("Kernel image path in this domain is not defined"));
+        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
+                       _("Kernel image path is not defined. With sev_snp=on, pass an igvm path"));
         return -1;
+    }
+
+    if (vmdef->sec &&
+        vmdef->sec->sectype == VIR_DOMAIN_LAUNCH_SECURITY_SEV_SNP) {
+        if (virJSONValueObjectAppendString(payload, "igvm", vmdef->os.kernel) < 0)
+            return -1;
+
+        if (vmdef->sec->data.sev_snp.host_data) {
+            size_t len;
+            const size_t host_data_len = 32;
+            g_autofree unsigned char *buf = NULL;
+            g_autofree char *host_data = NULL;
+
+            /* Libvirt provided host_data is base64 encoded and cloud-hypervisor
+               requires host_data as hex encoded. Base64 decode and hex encode
+               before sending to cloud-hypervisor.*/
+            buf = g_base64_decode(vmdef->sec->data.sev_snp.host_data, &len);
+            if (len != host_data_len) {
+                virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
+                               _("Invalid host_data provided. Expected '%1$zu' bytes"),
+                               host_data_len);
+                return -1;
+            }
+
+            host_data = virStringFormatHex(buf, host_data_len);
+            if (virJSONValueObjectAppendString(payload, "host_data",
+                                               host_data) < 0)
+                return -1;
+        }
     } else {
-        if (virJSONValueObjectAppendString(payload, "kernel", vmdef->os.kernel) < 0)
-            return -1;
-    }
-
-    if (vmdef->os.cmdline) {
-        if (virJSONValueObjectAppendString(payload, "cmdline", vmdef->os.cmdline) < 0)
-            return -1;
-    }
-
-    if (vmdef->os.initrd != NULL) {
-        if (virJSONValueObjectAppendString(payload, "initramfs", vmdef->os.initrd) < 0)
+        if (virJSONValueObjectAdd(&payload,
+                                  "s:kernel", vmdef->os.kernel,
+                                  "S:cmdline", vmdef->os.cmdline,
+                                  "S:initramfs", vmdef->os.initrd,
+                                  NULL) < 0)
             return -1;
     }
 
     if (virJSONValueObjectAppend(content, "payload", &payload) < 0)
-    return -1;
+        return -1;
 
     return 0;
 }
@@ -278,6 +302,44 @@ virCHMonitorBuildDisksJson(virJSONValue *content, virDomainDef *vmdef)
     return 0;
 }
 
+static int
+virCHMonitorBuildRngJson(virJSONValue *content, virDomainDef *vmdef)
+{
+    g_autoptr(virJSONValue) rng = virJSONValueNewObject();
+
+    if (vmdef->nrngs == 0) {
+        return 0;
+    }
+
+    if (vmdef->rngs[0]->model != VIR_DOMAIN_RNG_MODEL_VIRTIO) {
+        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
+                       _("Only virtio model is supported for RNG devices"));
+        return -1;
+    }
+
+    switch (vmdef->rngs[0]->backend) {
+    case VIR_DOMAIN_RNG_BACKEND_RANDOM:
+        if (virJSONValueObjectAppendString(rng, "src", vmdef->rngs[0]->source.file) < 0)
+            return -1;
+
+        if (virJSONValueObjectAppend(content, "rng", &rng) < 0)
+            return -1;
+
+        break;
+
+    case VIR_DOMAIN_RNG_BACKEND_EGD:
+    case VIR_DOMAIN_RNG_BACKEND_BUILTIN:
+        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
+                       _("Only RANDOM backend is supported for RNG devices"));
+        return -1;
+
+    case VIR_DOMAIN_RNG_BACKEND_LAST:
+        break;
+    }
+
+    return 0;
+}
+
 /**
  * virCHMonitorBuildNetJson:
  * @net: pointer to a guest network definition
@@ -426,6 +488,23 @@ virCHMonitorBuildDevicesJson(virJSONValue *content,
     return 0;
 }
 
+static int
+virCHMonitorBuildPlatformJson(virJSONValue *content, virDomainDef *vmdef)
+{
+    if (vmdef->sec &&
+        vmdef->sec->sectype == VIR_DOMAIN_LAUNCH_SECURITY_SEV_SNP) {
+        g_autoptr(virJSONValue) platform = virJSONValueNewObject();
+
+        if (virJSONValueObjectAppendBoolean(platform, "sev_snp", 1) < 0)
+            return -1;
+
+        if (virJSONValueObjectAppend(content, "platform", &platform) < 0)
+            return -1;
+    }
+
+    return 0;
+}
+
 static int
 virCHMonitorBuildVMJson(virCHDriver *driver, virDomainDef *vmdef,
                         char **jsonstr)
@@ -454,10 +533,15 @@ virCHMonitorBuildVMJson(virCHDriver *driver, virDomainDef *vmdef,
             return -1;
     }
 
+    if (virCHMonitorBuildPlatformJson(content, vmdef) < 0)
+        return -1;
 
     if (virCHMonitorBuildDisksJson(content, vmdef) < 0)
         return -1;
 
+    if (virCHMonitorBuildRngJson(content, vmdef) < 0)
+        return -1;
+
     if (virCHMonitorBuildDevicesJson(content, vmdef) < 0)
         return -1;
 
@@ -541,10 +625,12 @@ chMonitorCreateSocket(const char *socket_path)
 virCHMonitor *
 virCHMonitorNew(virDomainObj *vm, virCHDriverConfig *cfg, int logfile)
 {
+    virCHDomainObjPrivate *priv = vm->privateData;
     g_autoptr(virCHMonitor) mon = NULL;
     g_autoptr(virCommand) cmd = NULL;
     int socket_fd = 0;
     int event_monitor_fd;
+    int rv;
 
     if (virCHMonitorInitialize() < 0)
         return NULL;
@@ -603,6 +689,7 @@ virCHMonitorNew(virDomainObj *vm, virCHDriverConfig *cfg, int logfile)
     virCommandSetErrorFD(cmd, &logfile);
     virCommandNonblockingFDs(cmd);
     virCommandSetUmask(cmd, 0x002);
+
     socket_fd = chMonitorCreateSocket(mon->socketpath);
     if (socket_fd < 0) {
         virReportSystemError(errno,
@@ -614,13 +701,26 @@ virCHMonitorNew(virDomainObj *vm, virCHDriverConfig *cfg, int logfile)
     virCommandAddArg(cmd, "--api-socket");
     virCommandAddArgFormat(cmd, "fd=%d", socket_fd);
     virCommandPassFD(cmd, socket_fd, VIR_COMMAND_PASS_FD_CLOSE_PARENT);
-
     virCommandAddArg(cmd, "--event-monitor");
     virCommandAddArgFormat(cmd, "path=%s", mon->eventmonitorpath);
+    virCommandSetPidFile(cmd, priv->pidfile);
+    virCommandDaemonize(cmd);
 
     /* launch Cloud-Hypervisor socket */
-    if (virCommandRunAsync(cmd, &mon->pid) < 0)
+    if (virCommandRun(cmd, NULL) < 0) {
+        VIR_DEBUG("CH vm=%p name=%s failed to spawn",
+                  vm, vm->def->name);
         return NULL;
+    }
+
+    if ((rv = virPidFileReadPath(priv->pidfile, &vm->pid)) < 0) {
+        virReportSystemError(-rv,
+                             _("Domain %1$s didn't show up"),
+                             vm->def->name);
+        return NULL;
+    }
+    VIR_DEBUG("CH vm=%p name=%s running with pid=%lld",
+              vm, vm->def->name, (long long)vm->pid);
 
     /* open the reader end of fifo before start Event Handler */
     while ((event_monitor_fd = open(mon->eventmonitorpath, O_RDONLY)) < 0) {
@@ -668,12 +768,6 @@ void virCHMonitorClose(virCHMonitor *mon)
     if (!mon)
         return;
 
-    if (mon->pid > 0) {
-        /* try cleaning up the Cloud-Hypervisor process */
-        virProcessAbort(mon->pid);
-        mon->pid = 0;
-    }
-
     if (mon->handle)
         curl_easy_cleanup(mon->handle);
 
@@ -798,6 +892,8 @@ virCHMonitorPutNoContent(virCHMonitor *mon, const char *endpoint,
     if (responseCode == 200 || responseCode == 204)
         ret = 0;
 
+    curl_slist_free_all(headers);
+
     return ret;
 }
 
@@ -843,6 +939,7 @@ virCHMonitorGet(virCHMonitor *mon, const char *endpoint, virJSONValue **response
 
  cleanup:
     g_free(data.content);
+    curl_slist_free_all(headers);
     /* reset the libcurl handle to avoid leaking a stack pointer to data */
     curl_easy_reset(mon->handle);
 
@@ -1083,7 +1180,7 @@ virCHMonitorBuildRestoreJson(virDomainDef *vmdef,
         g_autoptr(virJSONValue) nets = virJSONValueNewArray();
         for (i = 0; i < vmdef->nnets; i++) {
             g_autoptr(virJSONValue) net_json = virJSONValueNewObject();
-            g_autofree char *id = g_strdup_printf("%s_%ld", CH_NET_ID_PREFIX, i);
+            g_autofree char *id = g_strdup_printf("%s_%zu", CH_NET_ID_PREFIX, i);
             if (virJSONValueObjectAppendString(net_json, "id", id) < 0)
                 return -1;
             if (virJSONValueObjectAppendNumberInt(net_json, "num_fds", vmdef->nets[i]->driver.virtio.queues))
@@ -1152,10 +1249,7 @@ virCHMonitorGetIOThreads(virCHMonitor *mon,
             if (!(map = virProcessGetAffinity(iothreadinfo->iothread_id)))
                 goto error;
 
-            if (virBitmapToData(map, &(iothreadinfo->cpumap),
-                                &(iothreadinfo->cpumaplen)) < 0) {
-                goto error;
-            }
+            virBitmapToData(map, &(iothreadinfo->cpumap), &(iothreadinfo->cpumaplen));
 
             /* Append to iothreadinfolist */
             iothreadinfolist[niothreads] = g_steal_pointer(&iothreadinfo);
